/**
 * mock-xmlhttprequest v8.0.0
 * (c) 2022 Bertrand Guay-Paquet
 * @license ISC
 */
'use strict';

const Utils = require('./Utils.cjs');

/**
 * Mock server for responding to XMLHttpRequest mocks from the class MockXhr. Provides simple route
 * matching and request handlers to make test harness creation easier.
 */
class MockXhrServer {
    /**
     * Constructor
     *
     * @param xhrMock XMLHttpRequest mock class
     * @param routes Routes
     */
    constructor(xhrMock, routes) {
        this._MockXhr = xhrMock;
        this._requests = [];
        this._routes = {};
        if (routes) {
            Object.entries(routes).forEach(([method, [urlMatcher, handler]]) => {
                this.addHandler(method, urlMatcher, handler);
            });
        }
        xhrMock.onSend = (request) => { this._handleRequest(request); };
        // Setup a mock request factory for users
        this._xhrFactory = () => new this._MockXhr();
    }
    get MockXhr() {
        return this._MockXhr;
    }
    /**
     * For backwards compatibility with versions < 4.1.0
     *
     * @deprecated Use the MockXhr property instead
     */
    get xhrMock() {
        return this._MockXhr;
    }
    get xhrFactory() {
        return this._xhrFactory;
    }
    /**
     * Install the server's XMLHttpRequest mock in the global context. You can specify a different
     * context with the optional `context` argument. Revert with remove().
     *
     * @param context Context object (e.g. global, window)
     * @returns this
     */
    install(context = globalThis) {
        this._savedContext = context;
        // Distinguish between an undefined and a missing XMLHttpRequest property
        if ('XMLHttpRequest' in context) {
            this._savedContextHadXMLHttpRequest = true;
            this._savedXMLHttpRequest = context.XMLHttpRequest;
        }
        else {
            this._savedContextHadXMLHttpRequest = false;
        }
        context.XMLHttpRequest = this._MockXhr;
        return this;
    }
    /**
     * Revert the changes made by install(). Call this after your tests.
     */
    remove() {
        if (!this._savedContext) {
            throw new Error('remove() called without a matching install(context).');
        }
        if (this._savedContextHadXMLHttpRequest) {
            this._savedContext.XMLHttpRequest = this._savedXMLHttpRequest;
            delete this._savedXMLHttpRequest;
        }
        else {
            delete this._savedContext.XMLHttpRequest;
        }
        delete this._savedContext;
    }
    /**
     * Disable the effects of the timeout attribute on the XMLHttpRequest mock used by the server.
     */
    disableTimeout() {
        this._MockXhr.timeoutEnabled = false;
    }
    /**
     * Enable the effects of the timeout attribute on the XMLHttpRequest mock used by the server.
     */
    enableTimeout() {
        this._MockXhr.timeoutEnabled = true;
    }
    /**
     * Add a GET request handler.
     *
     * @param urlMatcher Url matcher
     * @param handler Request handler
     * @returns this
     */
    get(urlMatcher, handler) {
        return this.addHandler('GET', urlMatcher, handler);
    }
    /**
     * Add a POST request handler.
     *
     * @param urlMatcher Url matcher
     * @param handler Request handler
     * @returns this
     */
    post(urlMatcher, handler) {
        return this.addHandler('POST', urlMatcher, handler);
    }
    /**
     * Add a PUT request handler.
     *
     * @param urlMatcher Url matcher
     * @param handler Request handler
     * @returns this
     */
    put(urlMatcher, handler) {
        return this.addHandler('PUT', urlMatcher, handler);
    }
    /**
     * Add a DELETE request handler.
     *
     * @param urlMatcher Url matcher
     * @param handler Request handler
     * @returns this
     */
    delete(urlMatcher, handler) {
        return this.addHandler('DELETE', urlMatcher, handler);
    }
    /**
     * Add a request handler.
     *
     * @param method HTTP method
     * @param urlMatcher Url matcher
     * @param handler Request handler
     * @returns this
     */
    addHandler(method, urlMatcher, handler) {
        // Match the processing done in MockXHR for the method name
        method = Utils.normalizeHTTPMethodName(method);
        const routes = this._routes[method] ?? (this._routes[method] = []);
        routes.push({ urlMatcher, handler, count: 0 });
        return this;
    }
    /**
     * Set the default request handler for requests that don't match any route.
     *
     * @param handler Request handler
     * @returns this
     */
    setDefaultHandler(handler) {
        this._defaultRoute = { handler, count: 0 };
        return this;
    }
    /**
     * Return 404 responses for requests that don't match any route.
     *
     * @returns this
     */
    setDefault404() {
        return this.setDefaultHandler({ status: 404 });
    }
    /**
     * @returns Array of requests received by the server. Entries: { method, url, headers, body? }
     */
    getRequestLog() {
        return [...this._requests];
    }
    _handleRequest(request) {
        // Record the request for easier debugging
        this._requests.push({
            method: request.method,
            url: request.url,
            headers: request.requestHeaders.getHash(),
            body: request.body,
        });
        const route = this._findFirstMatchingRoute(request) ?? this._defaultRoute;
        if (route) {
            // Routes can have arrays of handlers. Each one is used once and the last one is used if out
            // of elements.
            let { handler } = route;
            if (Array.isArray(handler)) {
                handler = handler[Math.min(handler.length - 1, route.count)];
            }
            route.count += 1;
            if (typeof handler === 'function') {
                handler(request);
            }
            else if (handler === 'error') {
                request.setNetworkError();
            }
            else if (handler === 'timeout') {
                request.setRequestTimeout();
            }
            else {
                request.respond(handler.status, handler.headers, handler.body, handler.statusText);
            }
        }
    }
    _findFirstMatchingRoute(request) {
        const method = Utils.normalizeHTTPMethodName(request.method);
        if (!this._routes[method]) {
            return undefined;
        }
        const { url } = request;
        return this._routes[method].find((route) => {
            const { urlMatcher } = route;
            if (typeof urlMatcher === 'function') {
                return urlMatcher(url);
            }
            else if (urlMatcher instanceof RegExp) {
                return urlMatcher.test(url);
            }
            return urlMatcher === url;
        });
    }
}

module.exports = MockXhrServer;
//# sourceMappingURL=MockXhrServer.cjs.map
