import MockXhrRequest from './MockXhrRequest';
import RequestData from './RequestData';
import XhrEventTarget from './XhrEventTarget';
import type { MockXhrResponseReceiver } from './MockXhrResponseReceiver';
export declare type OnCreateCallback = (xhr: MockXhr) => void;
export declare type OnSendCallback = (this: MockXhrRequest, request: MockXhrRequest) => void;
/**
 * XMLHttpRequest mock for testing.
 * Based on https://xhr.spec.whatwg.org version '18 August 2020'.
 *
 * Supports:
 *  - Events and states
 *  - open(), setRequestHeader(), send() and abort()
 *  - Upload and download progress events
 *  - Response status, statusText, headers and body
 *  - The timeout attribute (can be disabled)
 *  - Simulating a network error (see setNetworkError())
 *  - Simulating a request timeout (see setRequestTimeout())
 *
 * Partial support:
 *  - overrideMimeType(): throws when required, but has no other effect.
 *  - responseType: '', 'text' and 'json' are fully supported. The responseType values have no
 *    effect on the response body passed to setResponseBody().
 *  - responseXml: the response body is not converted to a document response. To get a document
 *    response, pass it directly as the response body in setResponseBody().
 *  - responseUrl: the final request URL after redirects isn't automatically set. This can be
 *    emulated in a request handler.
 *
 * Not supported:
 * - Synchronous requests (i.e. async set to false in open())
 * - Parsing the request URL in open() and throwing SyntaxError on failure.
 */
export default class MockXhr extends XhrEventTarget implements XMLHttpRequest, MockXhrResponseReceiver {
    private _requestHeaders;
    private _method?;
    private _url?;
    private _readyState;
    private _timeout;
    private _withCredentials;
    private _currentRequest?;
    private readonly _upload;
    responseURL: string;
    private _responseType;
    private _response;
    private _sendFlag?;
    private _uploadListenerFlag?;
    private _uploadCompleteFlag?;
    private _timedOutFlag?;
    private _timeoutReference;
    private _timeoutTask;
    constructor();
    static readonly UNSENT = 0;
    static readonly OPENED = 1;
    static readonly HEADERS_RECEIVED = 2;
    static readonly LOADING = 3;
    static readonly DONE = 4;
    readonly UNSENT = 0;
    readonly OPENED = 1;
    readonly HEADERS_RECEIVED = 2;
    readonly LOADING = 3;
    readonly DONE = 4;
    onreadystatechange: ((this: XMLHttpRequest, ev: Event) => any) | null;
    /**
     * Per-instance flag to enable the effects of the timeout attribute
     */
    timeoutEnabled: boolean;
    /**
     * Global flag to enable the effects of the timeout attribute
     */
    static timeoutEnabled: boolean;
    /**
     * Hook for creation of MockXhr instances
     */
    static onCreate?: OnCreateCallback;
    /**
     * Per-instance hook for XMLHttpRequest.send(). Executes in an empty callstack.
     */
    onSend?: OnSendCallback;
    /**
     * Global hook for XMLHttpRequest.send(). Executes in an empty callstack.
     */
    static onSend?: OnSendCallback;
    get currentRequest(): MockXhrRequest | undefined;
    /**
     * @returns All response headers as an object. The header names are in lower-case.
     */
    getResponseHeadersHash(): Record<string, string>;
    /**
     * Fire a request upload progress event.
     *
     * @param request Originating request
     * @param transmitted Bytes transmitted
     */
    uploadProgress(request: RequestData, transmitted: number): void;
    /**
     * Set the response headers. Changes the request's readyState to HEADERS_RECEIVED.
     *
     * @param request Originating request
     * @param status Response http status (default 200)
     * @param headers Name-value headers (optional)
     * @param statusText Response http status text (optional)
     */
    setResponseHeaders(request: RequestData, status?: number, headers?: Record<string, string> | null, statusText?: string): void;
    /**
     * Fire a response progress event. Changes the request's readyState to LOADING.
     *
     * @param request Originating request
     * @param transmitted Transmitted bytes
     * @param length Total bytes
     */
    downloadProgress(request: RequestData, transmitted: number, length: number): void;
    /**
     * Set the response body. Changes the request's readyState to DONE.
     *
     * @param request Originating request
     * @param body Response body
     */
    setResponseBody(request: RequestData, body: any): void;
    /**
     * Simulate a network error. Changes the request's readyState to DONE.
     *
     * @param request Originating request
     */
    setNetworkError(request: RequestData): void;
    /**
     * Simulate a request timeout. Changes the request's readyState to DONE.
     *
     * @param request Originating request
     */
    setRequestTimeout(request: RequestData): void;
    /**
     * https://xhr.spec.whatwg.org/#dom-xmlhttprequest-readystate
     *
     * @returns Client's state
     */
    get readyState(): number;
    /**
     * Set the request method and url.
     * https://xhr.spec.whatwg.org/#the-open()-method
     *
     * @param method Request HTTP method (GET, POST, etc.)
     * @param url Request url
     * @param async Async request flag (only true is supported)
     */
    open(method: string, url: string | URL, async?: boolean): void;
    /**
     * Add a request header value.
     * https://xhr.spec.whatwg.org/#the-setrequestheader()-method
     *
     * @param name Header name
     * @param value Header value
     */
    setRequestHeader(name: string, value: string): void;
    /**
     * https://xhr.spec.whatwg.org/#dom-xmlhttprequest-timeout
     *
     * @returns timeout attribute
     */
    get timeout(): number;
    /**
     * https://xhr.spec.whatwg.org/#dom-xmlhttprequest-timeout
     *
     * @param value timeout value
     */
    set timeout(value: number);
    /**
     * https://xhr.spec.whatwg.org/#dom-xmlhttprequest-withcredentials
     *
     * @returns withCredentials attribute
     */
    get withCredentials(): boolean;
    /**
     * https://xhr.spec.whatwg.org/#dom-xmlhttprequest-withcredentials
     *
     * @param value withCredentials value
     */
    set withCredentials(value: boolean);
    /**
     * https://xhr.spec.whatwg.org/#the-upload-attribute
     *
     * @returns upload attribute
     */
    get upload(): XhrEventTarget;
    /**
     * Initiate the request.
     * https://xhr.spec.whatwg.org/#the-send()-method
     *
     * @param body Request body
     */
    send(body?: any): void;
    /**
     * Abort the request.
     * https://xhr.spec.whatwg.org/#the-abort()-method
     */
    abort(): void;
    /**
     * https://xhr.spec.whatwg.org/#dom-xmlhttprequest-status
     *
     * @returns status attribute
     */
    get status(): number;
    /**
     * https://xhr.spec.whatwg.org/#the-statustext-attribute
     *
     * @returns statusText attribute
     */
    get statusText(): string;
    /**
     * Get a response header value.
     * https://xhr.spec.whatwg.org/#dom-xmlhttprequest-getresponseheader
     *
     * @param name Header name
     * @returns Header value
     */
    getResponseHeader(name: string): string | null;
    /**
     * Get all response headers as a string.
     * https://xhr.spec.whatwg.org/#dom-xmlhttprequest-getallresponseheaders
     *
     * @returns Concatenated headers
     */
    getAllResponseHeaders(): string;
    /**
     * Throws when required, but has no other effect.
     * https://xhr.spec.whatwg.org/#dom-xmlhttprequest-overridemimetype
     *
     * @param mime MIME type
     */
    overrideMimeType(mime: string): void;
    /**
     * https://xhr.spec.whatwg.org/#dom-xmlhttprequest-responsetype
     *
     * @returns responseType attribute
     */
    get responseType(): XMLHttpRequestResponseType;
    /**
     * https://xhr.spec.whatwg.org/#dom-xmlhttprequest-responsetype
     *
     * @param value responseType value
     */
    set responseType(value: XMLHttpRequestResponseType);
    /**
     * https://xhr.spec.whatwg.org/#the-response-attribute
     *
     * @returns response attribute
     */
    get response(): any;
    /**
     * https://xhr.spec.whatwg.org/#the-responsetext-attribute
     *
     * @returns responseText attribute
     */
    get responseText(): string;
    /**
     * https://xhr.spec.whatwg.org/#dom-xmlhttprequest-responsexml
     *
     * @returns responseXML attribute
     */
    get responseXML(): any;
    /**
     * Note: the "process request body" task is in the MockXhr response methods
     * Process request end-of-body task. When the whole request is sent.
     * https://xhr.spec.whatwg.org/#the-send()-method
     *
     * @param bodySize request body size in bytes
     */
    private _requestEndOfBody;
    /**
     * Process response task. When the response headers are received.
     * https://xhr.spec.whatwg.org/#the-send()-method
     *
     * @param response Response
     */
    private _processResponse;
    /**
     * Handle response end-of-body for response.
     * https://xhr.spec.whatwg.org/#handle-response-end-of-body
     */
    private _handleResponseEndOfBody;
    /**
     * Handle errors for response.
     * https://xhr.spec.whatwg.org/#handle-errors
     */
    private _handleResponseErrors;
    /**
     * The request error steps for event 'event'.
     * https://xhr.spec.whatwg.org/#request-error-steps
     *
     * @param event Event name
     */
    private _requestErrorSteps;
    protected _callOnSend(onSend?: OnSendCallback): void;
    private _terminateRequest;
    private _fireEvent;
    private _fireUploadEvent;
    private _fireReadyStateChange;
    private _scheduleRequestTimeout;
    private _getPrototype;
}
//# sourceMappingURL=MockXhr.d.ts.map